package com.io.yutian.elementoriginlib.sql.api.action.query;

import com.io.yutian.elementoriginlib.sql.api.SQLAction;
import com.io.yutian.elementoriginlib.sql.api.SQLQuery;
import com.io.yutian.elementoriginlib.sql.api.function.SQLExceptionHandler;
import com.io.yutian.elementoriginlib.sql.api.function.SQLFunction;
import com.io.yutian.elementoriginlib.sql.api.function.SQLHandler;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.sql.SQLException;

/**
 * SQLQueryAction 是用于承载SQL查询语句并进行处理、返回并自动关闭连接的基本类。
 *
 * <ul>
 *     <li>同步执行 {@link #execute()}, {@link #execute(SQLFunction, SQLExceptionHandler)}
 *     <br>同步执行方法中有会抛出异常的方法与不抛出异常的方法，
 *     <br>若选择不抛出异常，则返回值可能为空，需要特殊处理。</li>
 *
 *     <li>异步执行 {@link #executeAsync(SQLHandler, SQLExceptionHandler)}
 *     <br>异步执行时将提供成功与异常两种处理方式
 *     <br>可自行选择是否对数据或异常进行处理
 *     <br>默认的异常处理器为 {@link #defaultExceptionHandler()}</li>
 * </ul>
 *
 * <b>注意： 无论是否异步，都不需要自行关闭ResultSet，本API已自动关闭</b>
 *
 * @author CarmJos
 * @since 0.2.6
 */
public interface QueryAction extends SQLAction<SQLQuery> {

    @Override
    @Contract("_,!null -> !null")
    default <R> @Nullable R executeFunction(@NotNull SQLFunction<@NotNull SQLQuery, R> function,
                                            @Nullable R defaultResult) throws SQLException {
        try (SQLQuery value = execute()) {
            R result = function.apply(value);
            return result == null ? defaultResult : result;
        }
    }

}
